from distutils.core import setup

setup(
    name='ab',
    author='Dan Masquelier',
    author_email='dancrew32+ab@gmail.com',
    url='https://github.com/dancrew32/ab/',
    version='0.1dev',
    packages=['ab'],
    license='Unlicense',
    description='Easy A/B Test Bucketing',
)
